package holeg.ui.model;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import holeg.model.AbstractCanvasObject;
import holeg.model.Edge;
import holeg.model.Flexibility;
import holeg.model.GroupNode;
import holeg.model.HolonElement;
import holeg.model.HolonObject;
import holeg.model.HolonSwitch;



/**
 * The Class Model is the class where everything is saved. All changes made to
 * the Data is managed via a controller.
 *
 * @author Gruppe14
 */
public class Model {

	private static final Logger log = Logger.getLogger(Model.class.getName());
    private int currentIteration = 0;
    private int maxIterations=100;
    
    /** the amount of iterations */
    
    /**
	 * All implemented FairnessModels:<br>
	 * {@link FairnessModel#MininumDemandFirst}<br>
	 * {@link FairnessModel#AllEqual}
	 */
	public enum FairnessModel{
		/**
		 * One Element of each HolonObject will be powered first, starting with the
		 * smallest Demand. If ale HolonObjects have an active Element, the
		 * simulation will try to fully supply as many HolonObjects as possible.
		 */
		MininumDemandFirst, 
		/**
		 * All HolonObjects will receive the same amount of energy.
		 */
		AllEqual
	}
    /** the Fairness model in use */
    private FairnessModel fairnessModel = FairnessModel.MininumDemandFirst;
   
    //TODO(Tom2021-12-20) remove replace with groupnode
    /*
     * Array of all CpsObjects in our canvas. It is set by default as an empty
     * list.
     */
    private ArrayList<AbstractCanvasObject> objectsOnCanvas;
    /*
     * Array of all CpsObjects in our canvas. It is set by default as an empty
     * list.
     */
    private ArrayList<Edge> edgesOnCanvas;
    

  

    /**
     * Constructor for the model. It initializes the categories and
     * objectsOnCanvas by default values. Listeners are also initialized by
     * default values.
     */
    public Model() {
    	log.fine("Init Model");
        setObjectsOnCanvas(new ArrayList<>());
        setEdgesOnCanvas(new ArrayList<>());
    }



    /**
     * Returns all Objects on the Canvas.
     *
     * @return the objectsOnCanvas
     */
    public ArrayList<AbstractCanvasObject> getObjectsOnCanvas() {
        return objectsOnCanvas;
    }

    /**
     * Sets all Objects on the Canvas.
     *
     * @param objectsOnCanvas the objectsOnCanvas to set
     */
    public void setObjectsOnCanvas(ArrayList<AbstractCanvasObject> objectsOnCanvas) {
        this.objectsOnCanvas = objectsOnCanvas;
    }

    /**
     * Get all Edges on the Canvas.
     *
     * @return the edgesOnCanvas
     */
    public ArrayList<Edge> getEdgesOnCanvas() {
        return edgesOnCanvas;
    }

    /**
     * Sets the edges on the Canvas.
     *
     * @param arrayList the edgesOnCanvas to set
     */
    public void setEdgesOnCanvas(ArrayList<Edge> arrayList) {
        this.edgesOnCanvas = arrayList;
    }

    /**
     * Adds an Edge to The Canvas.
     *
     * @param edge the edgesOnCanvas to add
     */
    public void addEdgeOnCanvas(Edge edge) {
        this.edgesOnCanvas.add(edge);
    }

    /**
     * Remove an edge from the Canvas.
     *
     * @param edge the edge to remove
     */
    public void removeEdgesOnCanvas(Edge edge) {
        this.edgesOnCanvas.remove(edge);
    }



    /**
     * Returns the maximum ITERATIONS.
     *
     * @return ITERATIONS
     */
    public int getMaxIterations() {
        return maxIterations;
    }


    /**
     * Returns the current iteration.
     *
     * @return current iteration
     */
    public int getCurrentIteration() {
        return currentIteration;
    }

    /**
     * sets the current Iteration.
     *
     * @param curIT the current Iteration
     */
    public void setCurrentIteration(int value) {
        this.currentIteration = value;
    }



    public List<HolonElement> getAllHolonElements() {
    	return getAllHolonObjectsOnCanvas().stream().flatMap(hO -> hO.getElements().stream()).collect(Collectors.toList());
    }
    
    public List<Flexibility> getAllFlexibilities() {
    	return getAllHolonObjectsOnCanvas().stream().flatMap(hO -> hO.getElements().stream().flatMap(ele -> ele.flexList.stream())).collect(Collectors.toList());
    }
    
    public void reset() {
    	resetFlexibilities();
    	resetEdges();
    }
    
    private void resetFlexibilities() {
    	getAllFlexibilities().forEach(flex -> flex.reset());
    }
    private void resetEdges() {
    	this.getEdgesOnCanvas().forEach(edge -> edge.reset());
    }
    
    
    public ArrayList<HolonObject> getAllHolonObjectsOnCanvas(){
		ArrayList<HolonObject> objectToReturn = new ArrayList<>();
		for(AbstractCanvasObject aCps : objectsOnCanvas) {
			if (aCps instanceof HolonObject hObject) {
				objectToReturn.add(hObject);
			}
    		if(aCps instanceof GroupNode groupnode){
    			objectToReturn.addAll(groupnode.getAllHolonObjectsRecursive().toList());
    		}
    	}
    	return objectToReturn;
    }
    

    
    public ArrayList<AbstractCanvasObject> getAllAbstractObjectsOnCanvas(){
		ArrayList<AbstractCanvasObject> objectToReturn = new ArrayList<AbstractCanvasObject>();
		getAllAsbtractObjectsRecursive(objectToReturn, getObjectsOnCanvas().stream());
    	return objectToReturn;
    }
    private void getAllAsbtractObjectsRecursive(ArrayList<AbstractCanvasObject> addObjectsToThisList, Stream<AbstractCanvasObject> listOfObjectsToSearch){
    	listOfObjectsToSearch.forEach(aCps -> {
    		if(aCps instanceof GroupNode groupnode){
    			getAllAsbtractObjectsRecursive(addObjectsToThisList, groupnode.getObjectsInThisLayer());
    		}
    		else{
    			addObjectsToThisList.add(aCps);
    		}
    	});
    }
    
    
    
    /**
     * get all Switches
     */
    public ArrayList<HolonSwitch> getAllSwitches() {
        ArrayList<HolonSwitch> switches = new ArrayList<>();
        for (AbstractCanvasObject obj : getObjectsOnCanvas()) {
            if (obj instanceof HolonSwitch sw) {
                switches.add(sw);
            } else if (obj instanceof GroupNode groupnode) {
            	switches.addAll(groupnode.getAllSwitchObjectsRecursive().toList());
            }
        }
        return switches;
    }

	
	/**
	 * @param iterations the number of steps for this simulation
	 */
	public void setIterations(int iterations){
		this.maxIterations=iterations;
	}
	
	/**
	 * @return the fairnessModel
	 */
	public FairnessModel getFairnessModel() {
		return fairnessModel;
	}
	
	/**
	 * @param fairnessModel the fairnessModel to set
	 */
	public void setFairnessModel(FairnessModel fairnessModel) {
		this.fairnessModel = fairnessModel;
	}

}
